@@ -72,7 +72,10 @@ ATTRIBUTE = 'attribute'
 REFMETHOD = 'meth'
 REFFUNCTION = 'func'
 INITPY = '__init__.py'
+# Regex expression for checking references of pattern like ":class:`~package_v1.module`"
 REF_PATTERN = ':(py:)?(func|class|meth|mod|ref|attr|exc):`~?[a-zA-Z0-9_\.<> ]*?`'
+# Regex expression for checking references of pattern like "~package_v1.subpackage.module"
+REF_PATTERN_LAST = '~(([a-zA-Z0-9_<>]*\.)*[a-zA-Z0-9_<>]*)'
 
 PROPERTY = 'property'
 
@@ -96,6 +99,8 @@ def build_init(app):
     app.env.docfx_signature_funcs_methods = {}
     # This store the uid-type mapping info
     app.env.docfx_info_uid_types = {}
+    # This stores uidnames of docstrings already parsed
+    app.env.docfx_uid_names = {}
 
     remote = getoutput('git remote -v')
 
@@ -184,23 +189,31 @@ def _refact_example_in_module_summary(lines):
     return new_lines
 
 
-def _resolve_reference_in_module_summary(lines):
+def _resolve_reference_in_module_summary(pattern, lines):
     new_lines = []
     for line in lines:
-        matched_objs = list(re.finditer(REF_PATTERN, line))
+        matched_objs = list(re.finditer(pattern, line))
         new_line = line
         for matched_obj in matched_objs:
             start = matched_obj.start()
             end = matched_obj.end()
             matched_str = line[start:end]
-            if '<' in matched_str and '>' in matched_str:
-                # match string like ':func:`***<***>`'
-                index = matched_str.index('<')
-                ref_name = matched_str[index+1:-2]
+            if pattern == REF_PATTERN:
+                if '<' in matched_str and '>' in matched_str:
+                    # match string like ':func:`***<***>`'
+                    index = matched_str.index('<')
+                    ref_name = matched_str[index+1:-2]
+                else:
+                    # match string like ':func:`~***`' or ':func:`***`'
+                    index = matched_str.index('~') if '~' in matched_str else matched_str.index('`')
+                    ref_name = matched_str[index+1:-1]
             else:
-                # match string like ':func:`~***`' or ':func:`***`'
-                index = matched_str.index('~') if '~' in matched_str else matched_str.index('`')
-                ref_name = matched_str[index+1:-1]
+                index = matched_str.rfind('.') + 1
+                if index == 0:
+                    # If there is no dot, push index to not include tilde
+                    index = 1
+                # Find the last component of the target. "~Queue.get" only returns <xref:get>
+                ref_name = matched_str[index:]
             new_line = new_line.replace(matched_str, '<xref:{}>'.format(ref_name))
         new_lines.append(new_line)
     return new_lines
@@ -270,6 +283,172 @@ def _extract_signature(obj_sig):
     return signature, parameters
 
 
+# Given documentation docstring, parse them into summary_info.
+def _extract_docstring_info(summary_info, summary, name):
+    top_summary = ""
+
+    # Initialize known types needing further processing.
+    var_types = {
+        ':rtype:': 'returns',
+        ':returns:': 'returns',
+        ':type': 'variables',
+        ':param': 'variables',
+        ':raises': 'exceptions',
+        ':raises:': 'exceptions'
+    }
+    
+    initial_index = -1
+        
+    # Prevent GoogleDocstring crashing on custom types and parse all xrefs to normal
+    if '~' in summary or '<xref:' in summary:
+        type_pairs = []
+        # Find first character after one of the three combination
+        initial_index = min(
+          max(0, summary.find('~')), 
+          max(0, summary.find('<xref'))
+        )
+
+        summary_part = summary[initial_index:]
+       
+        # Remove all occurrences of "~xref" and "<xref:type>"
+        while '~' in summary_part or "<xref:" in summary_part:
+
+            # Expecting format of "~xref"
+            if '~' in summary_part:
+                initial_index += summary_part.find('~')
+                original_type = summary[initial_index:initial_index+(summary[initial_index:].find(':'))]
+                initial_index += len(original_type)
+                original_type = " ".join(filter(None, re.split(r'\n|  |\|\s|\t', original_type)))
+                safe_type = original_type[1:]
+
+            # Expecting format of "<xref:type>:"
+            elif "<xref:" in summary_part:
+                initial_index += summary_part.find("<xref")
+                original_type = summary[initial_index:initial_index+(summary[initial_index:].find('>'))+1]
+                initial_index += len(original_type)
+                original_type = " ".join(filter(None, re.split(r'\n|  |\|\s|\t', original_type)))
+                safe_type = original_type[6:-1]
+            else:
+                raise ValueError("Encountered unexpected type in Exception docstring.")
+
+            type_pairs.append([original_type, safe_type])
+            summary_part = summary[initial_index:]
+
+        # Replace all the found occurrences
+        for pairs in type_pairs:
+            original_type, safe_type = pairs[0], pairs[1]
+            summary = summary.replace(original_type, safe_type)
+        
+    # Clean the string by cleaning newlines and backlashes, then split by white space.
+    config = Config(napoleon_use_param=True, napoleon_use_rtype=True)
+    # Convert Google style to reStructuredText
+    parsed_text = str(GoogleDocstring(summary, config))
+    
+    # Revert back to original type
+    if initial_index > -1:
+        for pairs in type_pairs:
+            original_type, safe_type = pairs[0], pairs[1]
+            parsed_text = parsed_text.replace(safe_type, original_type)
+
+    # Trim the top summary but maintain its formatting.
+    indexes = []
+    for types in var_types:
+        # Ensure that we look for exactly the string we want.
+        # Adding the extra space for non-colon ending types
+        # helps determine if we simply ran into desired occurrence
+        # or if we ran into a similar looking syntax but shouldn't
+        # parse upon it.
+        types += ' ' if types[-1] != ':' else ''
+        if types in parsed_text:
+            index = parsed_text.find(types)
+            if index > -1:
+                # For now, skip on parsing custom fields like attribute
+                if types == ':type ' and 'attribute::' in parsed_text:
+                    continue
+                indexes.append(index)
+
+    # If we found types needing further processing, locate its index,
+    # if we found empty array for indexes, stop processing further.
+    index = min(indexes) if indexes else 0
+
+    # Store the top summary separately.
+    if index == 0:
+        return summary
+
+    top_summary = parsed_text[:index]
+    parsed_text = parsed_text[index:]
+
+    # Clean up whitespace and other characters
+    parsed_text = " ".join(filter(None, re.split(r'\|\s', parsed_text))).split()
+
+    cur_type = ''
+    words = []
+    arg_name = ''
+    index = 0
+    # Used to track return type and description
+    r_type, r_descr = '', ''
+
+    # Using counter iteration to easily extract names rather than
+    # coming up with more complicated stopping logic for each tags.
+    while index <= len(parsed_text):
+        word = parsed_text[index] if index < len(parsed_text) else ""
+        # Check if we encountered specific words.
+        if word in var_types or index == len(parsed_text):               
+            # Finish processing previous section.
+            if cur_type:
+                if cur_type == ':type':
+                    summary_info[var_types[cur_type]][arg_name]['var_type'] = " ".join(words)
+                elif cur_type == ':param':
+                    summary_info[var_types[cur_type]][arg_name]['description'] = " ".join(words)
+                elif ":raises" in cur_type:
+                    summary_info[var_types[cur_type]].append({
+                        'var_type': arg_name,
+                        'description': " ".join(words)
+                    })
+                else:
+                    if cur_type == ':rtype:':
+                        r_type = " ".join(words)
+                    else:
+                        r_descr = " ".join(words)
+                    if r_type and r_descr:
+                        summary_info[var_types[cur_type]].append({
+                            'var_type': r_type,
+                            'description': r_descr
+                        })
+                        r_type, r_descr = '', ''
+
+            else:
+
+                # If after we processed the top summary and get in this state,
+                # likely we encountered a type that's not covered above or the docstring
+                # was formatted badly. This will likely break docfx job later on, should not
+                # process further.
+                if word not in var_types:
+                    raise ValueError(f"Encountered wrong formatting, please check docstring for {name}")
+   
+            # Reached end of string, break after finishing processing
+            if index == len(parsed_text):
+                break
+    
+            # Start processing for new section
+            cur_type = word
+            if cur_type in [':type', ':param', ':raises', ':raises:']:
+                index += 1
+                arg_name = parsed_text[index][:-1]
+                # Initialize empty dictionary if it doesn't exist already
+                if arg_name not in summary_info[var_types[cur_type]] and ':raises' not in cur_type:
+                    summary_info[var_types[cur_type]][arg_name] = {}
+
+            # Empty target string
+            words = []
+        else:
+            words.append(word)
+    
+        index += 1
+
+    return top_summary
+
+
 def _create_datam(app, cls, module, name, _type, obj, lines=None):
     """
     Build the data structure for an autodoc class
@@ -291,108 +470,6 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
 
         return path
 
-    def _extract_docstring_info(summary_info, summary):
-        top_summary = ""
-
-        # Initialize known types needing further processing.
-        var_types = {
-            ':rtype:': 'returns',
-            ':returns:': 'returns',
-            ':type': 'variables',
-            ':param': 'variables',
-            ':raises': 'exceptions',
-            ':raises:': 'exceptions'
-        }
-        
-        # Clean the string by cleaning newlines and backlashes, then split by white space.
-        config = Config(napoleon_use_param=True, napoleon_use_rtype=True)
-        # Convert Google style to reStructuredText
-        parsed_text = str(GoogleDocstring(summary, config))
-
-        # Trim the top summary but maintain its formatting.
-        indexes = []
-        for types in var_types:
-            index = parsed_text.find(types)
-            if index > -1:
-                # For now, skip on parsing custom fields like attribute
-                if types == ':type' and 'attribute::' in parsed_text:
-                    continue
-                indexes.append(index)
-
-        # If we found types needing further processing, locate its index,
-        # if we found empty array for indexes, stop processing further.
-        index = min(indexes) if indexes else 0
-
-        # Store the top summary separately.
-        if index == 0:
-            top_summary = summary
-        else:
-            top_summary = parsed_text[:index]
-            parsed_text = parsed_text[index:]
-
-            # Clean up whitespace and other characters
-            parsed_text = " ".join(filter(None, re.split(r'\n|  |\|\s', parsed_text))).split(" ")
-
-            cur_type = ''
-            words = []
-            arg_name = ''
-            index = 0
-
-            # Using counter iteration to easily extract names rather than
-            # coming up with more complicated stopping logic for each tags.
-            while index <= len(parsed_text):
-                word = parsed_text[index] if index < len(parsed_text) else ""
-                # Check if we encountered specific words.
-                if word in var_types or index == len(parsed_text):
-                    # Finish processing previous section.
-                    if cur_type:
-                        if cur_type == ':type':
-                            summary_info[var_types[cur_type]][arg_name]['var_type'] = " ".join(words)
-                        elif cur_type == ':param':
-                            summary_info[var_types[cur_type]][arg_name]['description'] = " ".join(words)
-                        elif ":raises" in cur_type:
-                            summary_info[var_types[cur_type]].append({
-                                'var_type': arg_name,
-                                'description': " ".join(words)
-                            })
-                        elif cur_type == ':rtype:':
-                            arg_name = " ".join(words)
-                        else:
-                            summary_info[var_types[cur_type]].append({
-                                'var_type': arg_name,
-                                'description': " ".join(words)
-                            })
-                    else:
-                        # If after we processed the top summary and get in this state,
-                        # likely we encountered a type that's not covered above or the docstring
-                        # was formatted badly. This will likely break docfx job later on, should not
-                        # process further.
-                        if word not in var_types:
-                            raise ValueError("Encountered wrong formatting, please check docstrings")
-   
-                    # Reached end of string, break after finishing processing
-                    if index == len(parsed_text):
-                        break
-    
-                    # Start processing for new section
-                    cur_type = word
-                    if cur_type in [':type', ':param', ':raises', ':raises:']:
-                        index += 1
-                        arg_name = parsed_text[index][:-1]
-                        # Initialize empty dictionary if it doesn't exist already
-                        if arg_name not in summary_info[var_types[cur_type]] and ':raises' not in cur_type:
-                            summary_info[var_types[cur_type]][arg_name] = {}
-
-                    # Empty target string
-                    words = []
-                else:
-                    words.append(word)
-    
-                index += 1
-
-        return top_summary
-
-
     if lines is None:
         lines = []
     short_name = name.split('.')[-1]
@@ -421,7 +498,6 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
                 lines = lines.split("\n") if lines else []
             except TypeError as e:
                 print("couldn't getdoc from method, function: {}".format(e))
-
         
         elif _type in [PROPERTY]:
             lines = inspect.getdoc(obj)
@@ -498,14 +574,22 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
 
     # Add extracted summary
     if lines != []:
-        lines = _resolve_reference_in_module_summary(lines)
+        # Resolve references for xrefs in two different formats.
+        # REF_PATTERN checks for patterns like ":class:`~google.package.module`"
+        lines = _resolve_reference_in_module_summary(REF_PATTERN, lines)
+        # REF_PATTERN_LAST checks for patterns like "~package.module"
+        lines = _resolve_reference_in_module_summary(REF_PATTERN_LAST, lines)
         summary = app.docfx_transform_string('\n'.join(_refact_example_in_module_summary(lines)))
     
         # Extract summary info into respective sections.
         if summary:
-            top_summary = _extract_docstring_info(summary_info, summary)
+            top_summary = _extract_docstring_info(summary_info, summary, name)
             datam['summary'] = top_summary
 
+    # If there is no summary, add a short snippet.
+    else:
+        datam['summary'] = "API documentation for {} {}.".format(name, _type)
+
     if args or sig or summary_info:
         datam['syntax'] = {}
 
@@ -548,8 +632,15 @@ def process_docstring(app, _type, name, obj, options, lines):
     """
     This function takes the docstring and indexes it into memory.
     """
-    # Use exception as class
 
+    # Check if we already processed this docstring.
+    if name in app.env.docfx_uid_names:
+        return
+
+    # Register current docstring to a set.
+    app.env.docfx_uid_names[name] = ''
+
+    # Use exception as class
     if _type == EXCEPTION:
         _type = CLASS
 
@@ -625,6 +716,7 @@ def insert_children_on_module(app, _type, datam):
 
     if MODULE not in datam or datam[MODULE] not in app.env.docfx_yaml_modules:
         return
+
     insert_module = app.env.docfx_yaml_modules[datam[MODULE]]
     # Find the module which the datam belongs to
     for obj in insert_module:
@@ -637,7 +729,6 @@ def insert_children_on_module(app, _type, datam):
             # If it is a function, add this to its module. No need for class and module since this is
             # done before calling this function.
             insert_module.append(datam)
-
             obj['references'].append(_create_reference(datam, parent=obj['uid']))
             break
         # Add classes & exceptions to module
@@ -718,9 +809,16 @@ def find_unique_name(package_name, entries):
         if name != "google" and name != "cloud" and entries[name] == 1:
             return [name, package_name[-1]]
 
+    # If there is no way to disambiguate, return the identifier name
+    return [package_name[-1]]
+
 # Used to disambiguate names that have same entries.
+# Returns a dictionary of names that are disambiguated in the form of:
+# {uidname: disambiguated_name}
 def disambiguate_toc_name(toc_yaml):
     name_entries = {}
+    disambiguated_names = {}
+
     for module in toc_yaml:
         module_name = module['name']
         if module_name not in name_entries:
@@ -739,13 +837,17 @@ def disambiguate_toc_name(toc_yaml):
             name_entries[module_name][module_name] = 1
 
         if 'items' in module:
-            disambiguate_toc_name(module['items'])
+            # Update the dictionary of dismabiguated names
+            disambiguated_names.update(disambiguate_toc_name(module['items']))
 
     for module in toc_yaml:
         module_name = module['name']
         # Check if there are multiple entires of module['name'], disambiguate if needed.
         if name_entries[module_name][module_name] > 1:
             module['name'] = ".".join(find_unique_name(module['uidname'].split("."), name_entries[module_name]))
+            disambiguated_names[module['uidname']] = module['name']
+
+    return disambiguated_names
 
 def build_finished(app, exception):
     """
@@ -797,6 +899,9 @@ def build_finished(app, exception):
     # caused by Windows case insensitive file system
     file_name_set = set()
 
+    # Used to disambiguate entry names
+    yaml_map = {}
+
     # Order matters here, we need modules before lower level classes,
     # so that we can make sure to inject the TOC properly
     for data_set in (app.env.docfx_yaml_modules,
@@ -926,33 +1031,7 @@ def build_finished(app, exception):
                     obj['source']['remote']['repo'] == 'https://apidrop.visualstudio.com/Content%20CI/_git/ReferenceAutomation'):
                         del(obj['source'])
 
-            # Output file
-            if uid.lower() in file_name_set:
-                filename = uid + "(%s)" % app.env.docfx_info_uid_types[uid]
-            else:
-                filename = uid
-
-            out_file = os.path.join(normalized_outdir, '%s.yml' % filename)
-            ensuredir(os.path.dirname(out_file))
-            if app.verbosity >= 1:
-                app.info(bold('[docfx_yaml] ') + darkgreen('Outputting %s' % filename))
-
-            with open(out_file, 'w') as out_file_obj:
-                out_file_obj.write('### YamlMime:UniversalReference\n')
-                try:
-                    dump(
-                        {
-                            'items': yaml_data,
-                            'references': references,
-                            'api_name': [],  # Hack around docfx YAML
-                        },
-                        out_file_obj,
-                        default_flow_style=False
-                    )
-                except Exception as e:
-                    raise ValueError("Unable to dump object\n{0}".format(yaml_data)) from e
-
-            file_name_set.add(filename)
+            yaml_map[uid] = [yaml_data, references]
            
             # Parse the name of the object.
             # Some types will need additional parsing to de-duplicate their names and contain
@@ -994,7 +1073,7 @@ def build_finished(app, exception):
         raise RuntimeError("No documentation for this module.")
 
     # Perform additional disambiguation of the name
-    disambiguate_toc_name(toc_yaml)
+    disambiguated_names = disambiguate_toc_name(toc_yaml)
 
     # Keeping uidname field carrys over onto the toc.yaml files, we need to
     # be keep using them but don't need them in the actual file
@@ -1036,7 +1115,7 @@ def build_finished(app, exception):
                     'langs': ['python'],
                     'type': 'package',
                     'kind': 'distribution',
-                    'summary': '',
+                    'summary': 'Reference API documentation for `{}`.'.format(app.config.project),
                     'children': index_children
                 }],
                 'references': index_references
@@ -1045,6 +1124,45 @@ def build_finished(app, exception):
             default_flow_style=False
         )
 
+    # Output files
+    for uid, data in iter(yaml_map.items()):
+
+        for yaml_item in data:
+            for obj in yaml_item:
+                # If the entry was disambiguated, update here:
+                obj_full_name = obj['fullName']
+                if disambiguated_names.get(obj_full_name):
+                    obj['name'] = disambiguated_names[obj_full_name]
+      
+        # data is formatted as [yaml_data, references]
+        yaml_data, references = data
+
+        if uid.lower() in file_name_set:
+            filename = uid + "(%s)" % app.env.docfx_info_uid_types[uid]
+        else:
+            filename = uid
+
+        out_file = os.path.join(normalized_outdir, '%s.yml' % filename)
+        ensuredir(os.path.dirname(out_file))
+        if app.verbosity >= 1:
+            app.info(bold('[docfx_yaml] ') + darkgreen('Outputting %s' % filename))
+
+        with open(out_file, 'w') as out_file_obj:
+            out_file_obj.write('### YamlMime:UniversalReference\n')
+            try:
+                dump(
+                    {
+                        'items': yaml_data,
+                        'references': references,
+                        'api_name': [],  # Hack around docfx YAML
+                    },
+                    out_file_obj,
+                    default_flow_style=False
+                )
+            except Exception as e:
+                raise ValueError("Unable to dump object\n{0}".format(yaml_data)) from e
+
+        file_name_set.add(filename)
 
 def missing_reference(app, env, node, contnode):
     reftarget = ''
